import threading
from abc import ABCMeta, abstractmethod
from typing import Optional

from time import time

from messageflux.base_service import BaseService
from messageflux.utils import ObservableEvent


class LoopMetrics:
    """
    a class that holds metrics for a single server loop
    """
    def __init__(self, loop_duration: float, exception: Optional[Exception] = None):
        """

        :param loop_duration: the duration (in seconds) that it took the loop to run
        :param exception: the exception (if any) that the loop raised
        """
        self._loop_duration = loop_duration
        self._exception = exception

    @property
    def loop_duration(self) -> float:
        """
        :return: the duration (in seconds) that it took the loop to run
        """
        return self._loop_duration

    @property
    def exception(self) -> Optional[Exception]:
        """
        :return: the exception (if any) that the loop raised
        """
        return self._exception


class ServerLoopService(BaseService, metaclass=ABCMeta):
    """
    this is a base class for services that uses a 'loop' as their running method
    """

    def __init__(self, *,
                 duration_after_loop_success: float = 0,
                 duration_after_loop_failure: float = 0, **kwargs):
        """

        :param duration_after_loop_success: the duration (in seconds) to wait after successful run of the loop.
        :param duration_after_loop_failure: the duration (in seconds) to wait after failed run of the loop.
        :param kwargs: the init args for base classes
        """
        super().__init__(**kwargs)
        self._duration_after_loop_success = duration_after_loop_success
        self._duration_after_loop_failure = duration_after_loop_failure
        self._loop_ended_event: ObservableEvent[LoopMetrics] = ObservableEvent()

    @property
    def loop_ended_event(self) -> ObservableEvent[LoopMetrics]:
        """
        an event that is fired when a single server loop has ended
        """
        return self._loop_ended_event

    def _run_service(self, cancellation_token: threading.Event):
        """
        runs the server loop continuously, until the cancellation token is set.

        :param cancellation_token: the cancellation token for this service
        """
        while not cancellation_token.is_set():
            loop_exception = None
            start_time = time()
            try:
                self._server_loop(cancellation_token)
            except Exception as ex:
                self._logger.exception('Server loop raised an exception')
                loop_exception = ex

            loop_duration = time() - start_time
            self._loop_ended_event.fire(LoopMetrics(loop_duration=loop_duration, exception=loop_exception))

            wait_duration = self._duration_after_loop_success
            if loop_exception is not None:
                wait_duration = self._duration_after_loop_failure

            if wait_duration > 0:
                cancellation_token.wait(wait_duration)

    @abstractmethod
    def _server_loop(self, cancellation_token: threading.Event):
        """
        this method performs a single server loop. it should be implemented by the child classes

        :param cancellation_token: the cancellation token for this service
        """
        pass



import logging
import signal
import threading
from abc import ABCMeta, abstractmethod
from enum import Enum, unique
from typing import Optional

from messageflux.utils import ObservableEvent


@unique
class ServiceState(Enum):
    """
    the states of the service:
    INITIALIZING->(start)->STARTING->(prepare_service)->STARTED->(run_service)->STOPPING->(finalize_service)->STOPPED
    """
    INITIALIZED = "INITIALIZED"
    STARTING = 'STARTING'
    STARTED = 'STARTED'
    STOPPING = 'STOPPING'
    STOPPED = 'STOPPED'


class BaseService(metaclass=ABCMeta):
    """
    this class is the base class for all services
    """

    def __init__(self, *,
                 name: str = None,
                 should_stop_on_signal=True):
        """

        :param name: the name of this service. if None, the name of the type will be used
        :param should_stop_on_signal: if True, the service will try to register SIGTERM and SIGINT on stop method.
        """
        self._should_stop_on_signal: bool = should_stop_on_signal
        self._cancellation_token: threading.Event = threading.Event()
        self._cancellation_token.set()  # service starts as not_running
        self._logger = logging.getLogger(__name__)
        self._state_changed_event: ObservableEvent[ServiceState] = ObservableEvent()
        self._service_state = ServiceState.INITIALIZED

        if not name:
            name = type(self).__name__

        self._name = name

    @property
    def name(self) -> str:
        """
        the name of the service
        """
        return self._name

    @property
    def is_alive(self) -> bool:
        """
        this property is True, if the service is running.
        """
        return self._is_alive()

    @property
    def service_state(self) -> ServiceState:
        """
        this is the current service state
        """
        return self._service_state

    @property
    def state_changed_event(self) -> ObservableEvent[ServiceState]:
        """
        this is an Event, that can be used to register on server state changes
        """
        return self._state_changed_event

    def _is_alive(self) -> bool:
        """
        this may be overridden by child classes to change the 'is_alive' behaviour
        """
        return not self._cancellation_token.is_set()

    def _set_service_state(self, new_service_state: ServiceState):
        old_service_state = self._service_state
        self._service_state = new_service_state

        if old_service_state != new_service_state:
            self._state_changed_event.fire(new_service_state)

    def start(self):
        """
        starts the service, and blocks until 'stop' is called
        """
        self._cancellation_token.clear()
        if self._should_stop_on_signal:
            self._register_signals()
        server_exception = None
        try:
            self._set_service_state(ServiceState.STARTING)
            self._logger.info(f"Starting {self._name}")
            self._prepare_service()
            self._set_service_state(ServiceState.STARTED)
            self._run_service(cancellation_token=self._cancellation_token)
            self._cancellation_token.wait()
        except Exception as ex:
            self._logger.exception(f'Service raised an exception: {str(ex)}')
            self._cancellation_token.wait()
            server_exception = ex

        self._set_service_state(ServiceState.STOPPING)
        self._finalize_service(exception=server_exception)
        self._set_service_state(ServiceState.STOPPED)

    def _register_signals(self):
        if threading.current_thread() is threading.main_thread():
            self._logger.info("Registering Terminate Signals...")
            signal.signal(signal.SIGINT, lambda s, f: self.stop())
            signal.signal(signal.SIGTERM, lambda s, f: self.stop())
        else:
            self._logger.warning("Service doesn't run on main thread - can't register signals")

    def _prepare_service(self):
        """
        this method may be implemented by child classes, to perform some initialization logic
        before actually starting the service
        """
        pass

    @abstractmethod
    def _run_service(self, cancellation_token: threading.Event):
        """
        this method should be implemented by child classes, and actually run the service

        :param cancellation_token: the cancellation token that will be set when 'stop' is called
        """
        pass

    def _finalize_service(self, exception: Optional[Exception] = None):
        """
        this method may be implemented by child classes, to perform some cleanup logic
        after the service has finished running.

        :param exception: the exception (if any) that _run_service raised
        """
        pass

    def stop(self):
        """
        stops the service (sets the cancellation token, so the service will stop gracefully)
        """
        self._set_service_state(ServiceState.STOPPING)
        self._cancellation_token.set()



from abc import ABCMeta, abstractmethod
from dataclasses import dataclass
from typing import List, Tuple, Optional

from messageflux.device_reader_service import DeviceReaderService
from messageflux.iodevices.base import InputDevice, OutputDeviceManager, ReadResult
from messageflux.iodevices.base.common import MessageBundle


@dataclass
class PipelineResult:
    """
    a result from pipeline handler
    """
    output_device_name: str
    message_bundle: MessageBundle


class PipelineHandlerBase(metaclass=ABCMeta):
    """
    a pipeline handler base class. used to handle a single message that passes through the pipeline
    """

    @abstractmethod
    def handle_message(self,
                       input_device: InputDevice,
                       message_bundle: MessageBundle) -> Optional[PipelineResult]:
        """
        Handles a message from an input device. and returns a tuple of the output device name, message and headers to
        send to.

        :param input_device: The input device that sent the message.
        :param message_bundle: The message that was received.

        :return: None if the message should not be sent to any output device.
        PipelineResult if a message should be sent to the output device with the given name.
        """
        pass


class FixedRouterPipelineHandler(PipelineHandlerBase):
    """
    A pipeline handler that routes all messages to a single output device. without any processing.
    """

    def __init__(self, output_device_name: str):
        self._output_device_name = output_device_name

    def handle_message(self,
                       input_device: InputDevice,
                       message_bundle: MessageBundle) -> Optional[PipelineResult]:
        """
        routes the message to the fixed output

        :param input_device: The input device that sent the message.
        :param message_bundle: The message that was received.

        :return: a pipeline result that has the input message bundle, and the fixed output device name
        """
        output_message_bundle = MessageBundle(message=message_bundle.message.copy(), device_headers={})
        return PipelineResult(output_device_name=self._output_device_name,
                              message_bundle=output_message_bundle)


class PipelineService(DeviceReaderService):
    """
    a service that uses a PipelineHandler object to process messages from input, and send to output
    """
    def __init__(self, *, output_device_manager: OutputDeviceManager, pipeline_handler: PipelineHandlerBase, **kwargs):
        super().__init__(**kwargs)
        self._output_device_manager = output_device_manager
        self._pipeline_handler = pipeline_handler

    def _handle_messages(self, batch: List[Tuple[InputDevice, ReadResult]]):
        for input_device, read_result in batch:
            pipeline_result = self._pipeline_handler.handle_message(input_device, read_result)
            if pipeline_result is not None:
                output_device = self._output_device_manager.get_output_device(pipeline_result.output_device_name)
                output_device.send_message(message=pipeline_result.message_bundle.message,
                                           device_headers=pipeline_result.message_bundle.device_headers)
